/*
 * ui_gtk.c: 
 * GTK+ user interface for Umix.
 *
 * Copyright (C) 2002 Sakari Lehtonen <sakari@ionstream.fi>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>

#include "umix_common.h"
#include "ui.h"
#include "ui_gtk_prefs.h"
#include "ui_gtk_common.h"

#include "pixmaps/umix_logo.h"
#include "pixmaps/speaker_left.h"
#include "pixmaps/speaker_right.h"

#define UMIX_UI_GTK_VERSION_STR "0.2"

/* TODO:
 *
 * This is beta code, kludge saatoa atm .. 
 *
 * Before 0.3:
 * - Saving configuration options to umixrc .. section [gtk]
 * - When toggling record sources, update_all updates, fix
 * - Block signals instead of using kludge
 * - Selecting which channels to show (partly done, disabled atm)
 *
 * Later:
 * - Option for hiding the record vboxes
 * - Selecting which channel the bal adj controls 
 * - Entering volumes manually with control-e or something
 * - Find a way of focusing single channels
 * - Multiple mixer support
 * - Maybe split the code .. menu stuff in menu.c etc
 * - Check the atexit(ui_gtk_exit) stuff .. 
 * - Perhaps find a better way to reference created widgets, now pointer 
 *   tables are used.
 * - Clean up the code
 * - Optimize stuff .. too much looping over various things
 *
 */

enum ADJUSTMENT_DEFS
{
	ADJ_BAL_MIN = -VOL_MAX,
	ADJ_BAL_MAX = VOL_MAX,
	ADJ_BAL_STEP_INCR = 1,
	ADJ_BAL_PAGE_INCR = VOL_MAX,
	ADJ_BAL_PAGE_SIZE = 0,
	ADJ_VOL_MIN = -VOL_MAX,
	ADJ_VOL_MAX = 0,
	ADJ_VOL_STEP_INCR = VOL_MAX/10,
	ADJ_VOL_PAGE_INCR = VOL_MAX/10,
	ADJ_VOL_PAGE_SIZE = 0,
	DEF_UPDATE_TIMEOUT = 1500
};

/* global variables */
GtkWidget *top_window;

/* the channel number that the balance adj controls */
static int balance_channel = 0;

static GtkAdjustment **volume_adj;
static GtkAdjustment *balance_adj;

static GtkWidget **rec_checkbuttons;

static GtkWidget *top_vbox;
static GtkWidget *statusbar;
static GtkWidget *channel_hbox;
static GtkWidget *balance_hbox;
static GtkWidget *menu_bar;
static GtkWidget *about_window = NULL;

/* TODO: remove, instead block signals */
static int kludge = 0;

gtk_prefs gtk_config =
{
	TRUE  /* update_all */
};

/* global function prototypes */
int ui_gtk_main(void);
void ui_gtk_exit(void);

/*
void prefs_update(void);
*/

/* internal function prototypes */
static GtkWidget *create_top_window(void);
static GtkWidget *create_menu(GtkAccelGroup *accel_group);
static GtkWidget *create_volume_vbox(void);
static GtkWidget *create_channel_hbox(void);
static GtkWidget *create_channels_hbox(void);
static GtkWidget *create_balance_hbox(void);
static GtkWidget *create_record_vbox(void);

static void push_statusbar(gint context_id, const char *fmt, ...);
static void init_stuff(void);
static void create_stuff(void);
static gint update_all(gpointer data);
static void update_record_checkbuttons(void);

/* callback functions */
static void cb_vol_adj_chg(GtkAdjustment *adj, gpointer data);
static void cb_bal_adj_chg(GtkAdjustment *adj, gpointer data);
static void cb_rec_toggled(GtkToggleButton *button, gpointer data);
static void cb_mute_toggle(GtkWidget *widget, gpointer data);
static gchar *cb_vol_format(GtkScale *scale, gdouble value);

static void cb_settings_save(GtkWidget *widget, gpointer *data);
static void cb_settings_load(GtkWidget *widget, gpointer *data);

static void cb_quit_func(GtkButton *button, gpointer *data);

static void cb_show_about(GtkWidget *widget, gpointer data);
static void cb_about_delete(GtkWidget *widget, gpointer data);

/* the menu */
GtkItemFactoryEntry menu_items[] =
{
	{ "/_File", NULL, NULL, 0, "<Branch>" },
	{ "/File/_Load", "<control>L", cb_settings_load, 0,
		"<StockItem>", GTK_STOCK_OPEN },
	{ "/File/_Save", "<control>S", cb_settings_save, 0,
		"<StockItem>", GTK_STOCK_SAVE },
	{ "/File/sep1", NULL, NULL, 0, "<Separator>" },
	{ "/File/_Quit", "<control>Q", cb_quit_func, 0, 
		"<StockItem>", GTK_STOCK_QUIT },
	{ "/_Settings", NULL, NULL, 0, "<Branch>" },
	{ "/Settings/_Preferences...", "<control>P", cb_show_prefs, 0, 
		"<StockItem>", GTK_STOCK_PREFERENCES  },
	{ "/M_ute", NULL, NULL, 0, "<Branch>" },
	{ "/Mute/_Mute All", "M", cb_mute_toggle, 0, "<CheckItem>" },
	{ "/_Help", NULL, NULL, 0, "<Branch>" },
	{ "/Help/sep2", NULL, NULL, 0, "<Separator>" },
	{ "/Help/About", NULL, cb_show_about, 0, "<Item>" }
};

ui_func gtk_func =
{
	ui_gtk_main,
	ui_gtk_exit,
};

ui_func *get_ui_func(void)
{
	return &gtk_func;
}

int ui_gtk_main(void)
{
	/* initialize the environment */
	gtk_init(NULL, NULL);

	mixer_set_curr(0);
	chan_set_curr(0);

	/* create widgets */
	top_window = create_top_window();
	/* the top vbox contains everything */
	top_vbox = gtk_vbox_new(NONHOMOGENEOUS, 0);
	gtk_container_add(GTK_CONTAINER(top_window), top_vbox);
	gtk_widget_show(top_vbox);

	init_stuff();
	create_stuff();
	
	/* hmm not sure if this is the right way to do this .. */
	/* add timeout to check if the mixer is updated and update the
	 * sliders */
	if( gtk_config.update_all == TRUE )
		gtk_timeout_add(DEF_UPDATE_TIMEOUT, update_all, NULL);

	/* show */
	gtk_widget_show(top_window);
	gtk_main();

	return 0;
}

/* TODO: rename, fix, and check */
static void create_stuff(void)
{
	GtkAccelGroup *accel_group;
	gint context_id;
	
	/* set up the accelerator group */
	accel_group = gtk_accel_group_new();
	gtk_window_add_accel_group(GTK_WINDOW(top_window), accel_group);
	
	/* create the menu bar */
	menu_bar = create_menu(accel_group);
	gtk_box_pack_start(GTK_BOX(top_vbox), menu_bar, NOEXPAND, NOFILL, 0);
	gtk_widget_show(menu_bar);
	
	/* the channel_hbox contains all the channel vboxes */
	channel_hbox = create_channels_hbox();
	gtk_box_pack_start(GTK_BOX(top_vbox), channel_hbox, EXPAND, FILL, 0);
	gtk_widget_show(channel_hbox);

	/* create the balance adjustment */
	balance_hbox = create_balance_hbox();
	gtk_box_pack_start(GTK_BOX(top_vbox), balance_hbox, NOEXPAND,
		NOFILL, 0);
	gtk_widget_show(balance_hbox);
	
	/* status bar stuff */
	statusbar = gtk_statusbar_new();
	gtk_statusbar_set_has_resize_grip(GTK_STATUSBAR(statusbar), FALSE);
	context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
		"main");
	push_statusbar(context_id, "%s :: %s", mixer_get_name(), 
		mixer_get_path());
	gtk_box_pack_start(GTK_BOX(top_vbox), statusbar, NOEXPAND,
		NOFILL, 0);
	gtk_widget_show(statusbar);
}

/* this is called when the preferences ok button is pressed */
/*
void prefs_update(void)
{
}
*/

/* TODO: rename, check, and FIX */
static void init_stuff(void)
{
	int numchan;

	numchan = mixer_get_num_chan();
	volume_adj = emalloc(sizeof(GtkAdjustment) * numchan);
	rec_checkbuttons = emalloc(sizeof(GtkWidget *) * numchan);
}

/* volume adjustment callback */
static void cb_vol_adj_chg(GtkAdjustment *adj, gpointer data)
{
	float vol;

	if( kludge == 1 )
		return;

	chan_set_curr(GPOINTER_TO_INT(data));
	/* convert to int so we convert the extra decimals to 00, eg.
	 * 49.6000 comes 49.0000. This is done because when we call
	 * update_all, it reads the value from the device again, and 
	 * sets the slider to 49.0000 from 49.6000, thus moving it a bit */
	vol = (int)ABS(adj->value);
	chan_set_vol(vol);
}

/* balance adjustment callback */
static void cb_bal_adj_chg(GtkAdjustment *adj, gpointer data)
{
	float bal;
	/*gint context_id;*/
	
	if( kludge == 1 )
		return;

	chan_set_curr(GPOINTER_TO_INT(data));
	bal = adj->value/100.0;
#ifdef UMIX_DEBUG
	errormsg("cb_bal_adj_chg: ch=%d value=%f get_bal=%f", 
		chan_get_curr(), adj->value, chan_get_bal());
#endif
	/*
	context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
		"bal_adj");
	gtk_statusbar_pop(GTK_STATUSBAR(statusbar), context_id);
	push_statusbar(context_id, "Balance for channel %s set to %1.f", 
		chan_get_name(), bal*VOL_MAX);
	*/
	chan_set_bal(bal);
}

static void update_record_checkbuttons(void)
{
	int i;
	int origchan;
	
	kludge = 1;
	origchan = chan_get_curr();
	for(i=0; i<mixer_get_num_chan(); i++)
	{
		chan_set_curr(i);
		if( chan_is_record() )
		{
			gtk_toggle_button_set_active(
				GTK_TOGGLE_BUTTON(rec_checkbuttons[i]),
				chan_is_recsrc());
		}
	}
	kludge = 0;
	chan_set_curr(origchan);
}

static void cb_rec_toggled(GtkToggleButton *button, gpointer data)
{
	if( kludge == 1 )
		return;
	
	chan_set_curr(GPOINTER_TO_INT(data));
#ifdef UMIX_DEBUG
	errormsg("cb_rec_toggled: data=%d state=%d curr_ch=%s", 
		GPOINTER_TO_INT(data), gtk_toggle_button_get_active(button),
		chan_get_name());
#endif
	chan_toggle_rec();
	update_record_checkbuttons();
}

/* TODO: check, esp multiple mixer support */
static void cb_mute_toggle(GtkWidget *widget, gpointer data)
{
	int i;
	int orig_chan;
	gint context_id;

	kludge = 1;
	orig_chan = chan_get_curr();
	for(i=0; i<mixer_get_num_chan(); i++)
	{
		chan_set_curr(i);
		if( chan_is_mute() == 0 )
			chan_set_mute(MUTE);
		else
			chan_set_mute(UNMUTE);

		/* adjust the slider */
		gtk_adjustment_set_value(volume_adj[i], 
			-chan_get_vol());
	}
	kludge = 0;
	context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
		"toggle_mute");
	gtk_statusbar_pop(GTK_STATUSBAR(statusbar), context_id);
	push_statusbar(context_id, "Mute");

	chan_set_curr(orig_chan);
}

/* update the volume sliders */
static gint update_all(gpointer data)
{
	int i;
	int orig_chan;

	/* if data is TRUE, force refresh sliders */
	if( mixer_get_status() == 1 || (int)data == TRUE )
	{
#ifdef UMIX_DEBUG
		errormsg("update_all actually updating");
#endif
		kludge = 1;
		orig_chan = chan_get_curr();
		for(i=0; i<mixer_get_num_chan(); i++)
		{
			chan_set_curr(i);
			gtk_adjustment_set_value(volume_adj[i], 
				-chan_get_vol());
			if( chan_is_record() )
			{
				gtk_toggle_button_set_active(
					GTK_TOGGLE_BUTTON(rec_checkbuttons[i]),
					chan_is_recsrc());
			}
		}
		/*
	   	chan_set_curr(balance_channel);
		gtk_adjustment_set_value(balance_adj, chan_get_bal()*100.0);
		*/
		chan_set_curr(orig_chan);
		kludge = 0;
	}

	return TRUE;
}

/* pushes stuff to the statusbar with varargs */
static void push_statusbar(gint context_id, const char *fmt, ...)
{
	va_list ap;
	char buf[STRLEN];

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);
	
	gtk_statusbar_push(GTK_STATUSBAR(statusbar), context_id, buf);
}

/* creates and returns the hbox that holds all widgets for one channel */
static GtkWidget *create_channel_hbox(void)
{
	GtkWidget *top_hbox;
	GtkWidget *top_vbox;
	GtkWidget *vol_vbox;
	GtkWidget *rec_vbox;
	GtkWidget *vsep;

	top_hbox = gtk_hbox_new(NONHOMOGENEOUS, 6);
	gtk_widget_show(top_hbox);

	top_vbox = gtk_vbox_new(NONHOMOGENEOUS, 6);
	gtk_box_pack_start(GTK_BOX(top_hbox), top_vbox, NOEXPAND, NOFILL, 0);
	gtk_widget_show(top_vbox);

	/* the channel vbox */
	vol_vbox = create_volume_vbox();
	gtk_box_pack_start(GTK_BOX(top_vbox), vol_vbox, NOEXPAND, NOFILL, 0);
	gtk_widget_show(vol_vbox);

	/* the record vbox */
	rec_vbox = create_record_vbox();
	gtk_box_pack_start(GTK_BOX(top_vbox), rec_vbox, NOEXPAND, NOFILL, 0);
	gtk_widget_show(rec_vbox);
	
	/* TODO: don`t draw the last vseparator */
	vsep = gtk_vseparator_new();
	gtk_box_pack_start(GTK_BOX(top_hbox), vsep, EXPAND, FILL, 0);
	gtk_widget_show(vsep);
	
	return top_hbox;
}

/* creates and returns the top level hbox that holds all channels */
static GtkWidget *create_channels_hbox(void)
{
	GtkWidget *top_hbox;
	GtkWidget *chan_hbox;
	int i;

	top_hbox = gtk_hbox_new(NONHOMOGENEOUS, 6);
	gtk_container_set_border_width(GTK_CONTAINER(top_hbox), 2);
	/* TODO: add option what channels to show .. like a list of
	 * what channels to show, and fix the last vseparator not to be 
	 * add following options to be selectable :
	 *
	 * - Show only channels that are not muted 
	 * etc ..
	 */
	
	/* create all individual channel hboxes .. 
	 * hboxes because they contain the vseparators between
	 * channels too, so individual channels can be hidden/shown easily */
	for(i=0; i<mixer_get_num_chan(); i++)
	{
		chan_set_curr(i);
		chan_hbox = create_channel_hbox();
		gtk_box_pack_start(GTK_BOX(top_hbox), chan_hbox, 
			NOEXPAND, NOFILL, 0);
		gtk_widget_show(chan_hbox);

		/* store a global pointer so we can hide/show the channels
		 * later */
		/*
		chan_hboxes[0][chan_count++] = chan_hbox;

		*/
	}

	return top_hbox;
}

static GtkWidget *create_balance_hbox(void)
{
	GtkWidget *top_hbox;
	GtkWidget *scale;
	GtkWidget *speaker_left;
	GtkWidget *speaker_right;
	GdkPixbuf *pixbuf;
	
	top_hbox = gtk_hbox_new(NONHOMOGENEOUS, 0);

	/* left speaker */
	pixbuf = gdk_pixbuf_new_from_inline(-1, speaker_left_png, FALSE, 0);
	speaker_left = gtk_image_new_from_pixbuf(pixbuf);
	gtk_box_pack_start(GTK_BOX(top_hbox), speaker_left, NOEXPAND, 
		NOFILL, 2);
	gtk_widget_show(speaker_left);

	/* init the balance values */
	chan_set_curr(balance_channel);
	balance_adj = GTK_ADJUSTMENT(gtk_adjustment_new(chan_get_bal()*VOL_MAX, 
				ADJ_BAL_MIN, ADJ_BAL_MAX, ADJ_BAL_STEP_INCR,
				ADJ_BAL_PAGE_INCR, ADJ_BAL_PAGE_SIZE));
	g_signal_connect(G_OBJECT(balance_adj), "value_changed",
		(GtkSignalFunc)cb_bal_adj_chg, (gpointer)balance_channel);
	scale = gtk_hscale_new(GTK_ADJUSTMENT(balance_adj));
	gtk_scale_set_draw_value(GTK_SCALE(scale), FALSE);
	gtk_box_pack_start(GTK_BOX(top_hbox), scale, EXPAND, FILL, 0);
	gtk_widget_show(scale);
	
	/* right speaker */
	pixbuf = gdk_pixbuf_new_from_inline(-1, speaker_right_png, FALSE, 0);
	speaker_right = gtk_image_new_from_pixbuf(pixbuf);
	gtk_box_pack_start(GTK_BOX(top_hbox), speaker_right, NOEXPAND, 
		NOFILL, 2);
	gtk_widget_show(speaker_right);

	return top_hbox;
}

static gchar *cb_vol_format(GtkScale *scale, gdouble value)
{
	return g_strdup_printf("%d", (int)ABS(value));
}

/* creates and returns the volume vbox for one channel */
static GtkWidget *create_volume_vbox(void)
{
	GtkWidget *top_vbox;
	GtkWidget *chan_label;
	GtkWidget *vol_label;
	GtkWidget *vol_scale;
	GtkWidget *hsep;
	int curr_chan = chan_get_curr();

	top_vbox = gtk_vbox_new(NONHOMOGENEOUS, 0);
	/* channel label */
	chan_label = gtk_label_new(chan_get_label());
	gtk_misc_set_alignment(GTK_MISC(chan_label), 0.5, 0);
	gtk_box_pack_start(GTK_BOX(top_vbox), chan_label, 
		NOEXPAND, NOFILL, 0);
	gtk_widget_show(chan_label);

	hsep = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(top_vbox), hsep, EXPAND, FILL, 0);
	gtk_widget_show(hsep);

	/* TODO: maybe add option to show channel icons */
	vol_label = gtk_label_new(_("Volume:"));
	gtk_misc_set_alignment(GTK_MISC(vol_label), 0.5, 0);
	gtk_box_pack_start(GTK_BOX(top_vbox), vol_label, NOEXPAND,
		NOFILL, 0);
	gtk_widget_show(vol_label);

	/* store all the volume adjustments global so we can update them
	 * when the volume changes  */
	volume_adj[curr_chan] = GTK_ADJUSTMENT(
				gtk_adjustment_new(-chan_get_vol(),
				ADJ_VOL_MIN, ADJ_VOL_MAX, ADJ_VOL_STEP_INCR,
				ADJ_VOL_PAGE_INCR, ADJ_VOL_PAGE_SIZE));
	g_signal_connect(G_OBJECT(volume_adj[curr_chan]),"value_changed",
		(GtkSignalFunc)cb_vol_adj_chg, (gpointer)chan_get_curr());

	vol_scale = gtk_vscale_new(GTK_ADJUSTMENT(volume_adj[curr_chan]));
	/* connect the signal to format the value so that it isn`t shown
	 * as negative */
	g_signal_connect(G_OBJECT(vol_scale), "format-value",
		(GtkSignalFunc)cb_vol_format, NULL);
	gtk_scale_set_draw_value(GTK_SCALE(vol_scale), TRUE);
	gtk_widget_set_size_request(vol_scale, 1, VOL_MAX);
	gtk_box_pack_start(GTK_BOX(top_vbox), vol_scale, EXPAND, FILL, 0);
	gtk_widget_show(vol_scale);

	return top_vbox;
}

/* creates and returns a recbutton box for one channel */
static GtkWidget *create_record_vbox(void)
{
	GtkWidget *top_vbox;
	GtkWidget *button_vbox;
	GtkWidget *hsep;
	GtkWidget *rec_checkbutton;

	top_vbox = gtk_vbox_new(NONHOMOGENEOUS, 0);

	button_vbox = gtk_vbox_new(NONHOMOGENEOUS, 0);
	hsep = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(top_vbox), hsep, EXPAND, FILL, 0);
	gtk_widget_show(hsep);

#ifdef UMIX_DEBUG
	errormsg("Creating record vbox, curr_chan=%d", chan_get_curr());
#endif
	/* the rec check button, if channel can be a recording source */
	if( chan_is_record() )
	{
		gtk_box_pack_start(GTK_BOX(top_vbox), button_vbox, 
			NOEXPAND, NOFILL, 0);
		/* when this is first called, rec_group is NULL */
		rec_checkbutton = gtk_check_button_new_with_label("Rec");
		rec_checkbuttons[chan_get_curr()] = rec_checkbutton;
		if( chan_is_recsrc() )
		{
#ifdef UMIX_DEBUG
			errormsg("toggling channel %s as rec", chan_get_name());
#endif
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(
					rec_checkbutton), TRUE); 
		}
		g_signal_connect(G_OBJECT(rec_checkbutton), 
			"toggled", (GtkSignalFunc)cb_rec_toggled, 
			(gpointer)chan_get_curr());
		gtk_box_pack_start(GTK_BOX(button_vbox), 
			rec_checkbutton, NOEXPAND, NOFILL, 0);
		gtk_widget_show(rec_checkbutton);

	}
	/* pad with empty .. hmm not sure about this */
	else
		gtk_box_pack_start(GTK_BOX(top_vbox), button_vbox, 
			EXPAND, FILL, 12);
	
	gtk_widget_show(button_vbox);

	return top_vbox;
}

/* creates and returns the top menu */
static GtkWidget *create_menu(GtkAccelGroup *accel_group)
{
	GtkWidget *menu;
	GtkItemFactory *item_fact;

	/* create the item factory */
	item_fact = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", 
		accel_group);
	gtk_item_factory_create_items(item_fact, NELEMS(menu_items), 
		menu_items, NULL);
	menu = gtk_item_factory_get_widget(item_fact, "<main>");

	return menu;
}

static void cb_settings_save(GtkWidget *widget, gpointer *data)
{
	int retval;
	gint context_id;
	
	context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
		"settings");
	gtk_statusbar_pop(GTK_STATUSBAR(statusbar), context_id);

	retval = settings_save(globalopts.cfgpath);
	if( (settings_save(globalopts.cfgpath) == UMIX_ESETERR ) )
		push_statusbar(context_id, UMIX_UI_SAVEFAIL_STRFMT,
			globalopts.cfgpath, strerror(errno));
	else
		push_statusbar(context_id, UMIX_UI_SAVEOK_STRFMT,
			globalopts.cfgpath);
}

static void cb_settings_load(GtkWidget *widget, gpointer *data)
{
	int retval;
	gint context_id;
	
	context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
		"settings");
	gtk_statusbar_pop(GTK_STATUSBAR(statusbar), context_id);

	retval = settings_load(globalopts.cfgpath, UMIX_SET_MIXER);
	if( retval == UMIX_ESETERR )
		push_statusbar(context_id, "%s: %s",
			globalopts.cfgpath, strerror(errno));
	else if( retval == UMIX_ESETINVALID )
		push_statusbar(context_id, UMIX_UI_EINVALID_STR);
	else if( retval == UMIX_ESETNOTFOUND )
		push_statusbar(context_id, UMIX_UI_ENOTFOUND_STR);
	else if( retval == UMIX_ESETBINARY )
		push_statusbar(context_id, UMIX_UI_ESETBINARY_STR);
	else
		push_statusbar(context_id, UMIX_UI_LOADOK_STRFMT,
			globalopts.cfgpath);

	if( retval == 0 )
		update_all((int *)TRUE);
}

static void cb_about_delete(GtkWidget *widget, gpointer data)
{
	gtk_widget_destroy(about_window);
	about_window = NULL;
}       

static void cb_show_about(GtkWidget *widget, gpointer data)
{
	GtkWidget *frame;
	GtkWidget *vbox;
	GtkWidget *label;
	GtkWidget *logo;
	GdkPixbuf *pixbuf;
	char buf[BUFLEN];

	if (about_window != NULL)
	{
		gdk_window_raise(about_window->window);
		return;
	}

	about_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_wmclass(GTK_WINDOW(about_window), "about", "Umix");
	g_signal_connect(G_OBJECT(about_window), "delete_event", 
		G_CALLBACK(cb_about_delete), NULL);

	gtk_window_set_title(GTK_WINDOW(about_window), "About - Umix");
	gtk_window_set_resizable(GTK_WINDOW(about_window), FALSE);
	gtk_window_set_position(GTK_WINDOW(about_window), GTK_WIN_POS_MOUSE);
	gtk_container_set_border_width(GTK_CONTAINER(about_window), 5);

	frame = gtk_frame_new(NULL);
	gtk_container_add(GTK_CONTAINER(about_window), frame);
	gtk_widget_show(frame);

	vbox = gtk_vbox_new(NONHOMOGENEOUS, 10);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
	gtk_container_add(GTK_CONTAINER(frame),vbox);
	gtk_widget_show(vbox);

	pixbuf = gdk_pixbuf_new_from_inline(-1, umix_logo_png, FALSE, 0);
	logo = gtk_image_new_from_pixbuf(pixbuf);
	gtk_box_pack_start(GTK_BOX(vbox), logo, NOEXPAND, NOFILL, 0);
	gtk_widget_show(logo);

	snprintf(buf, sizeof(buf), 
		"%s - GTK+ UI v%s\n\n"
		_("A program for adjusting soundcard volumes\n"
			"and other features in soundcard mixers.\n\n"
			"Copyright (C) 2002 Sakari Lehtonen\n"
			"http://www.ionstream.fi/sakari/umix/\n")
		, get_version_str(), UMIX_UI_GTK_VERSION_STR);

	label = gtk_label_new(buf);
	gtk_box_pack_start(GTK_BOX(vbox), label, NOEXPAND, NOFILL, 0);
	gtk_widget_show(label);

	gtk_widget_show(about_window);
}

/* creates and returns the top window */
static GtkWidget *create_top_window(void)
{
	char window_title[LINEWIDTH];
	GtkWidget *window;

	/* create window */
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	/* set attributes */
	snprintf(window_title, sizeof(window_title), "%s", 
		get_version_str()); 
	gtk_window_set_title(GTK_WINDOW(window), window_title);
	gtk_container_set_border_width(GTK_CONTAINER(window), NOSPACING);
	gtk_window_set_resizable(GTK_WINDOW(window), FALSE);
	gtk_window_set_wmclass(GTK_WINDOW(window), "umix", "Umix");

	/* connect signals */
	g_signal_connect(G_OBJECT(window), "destroy",
		G_CALLBACK(cb_quit_func), NULL);

	return window;
}

static void cb_quit_func(GtkButton *button, gpointer *data)
{
	gtk_main_quit();
}

void ui_gtk_exit(void)
{
#ifdef UMIX_DEBUG
	errormsg("ui_gtk_exit: GTK UI deinitialized");
#endif
	free(rec_checkbuttons);
	free(volume_adj);
}
