/*
 * ui_ncurses.c: 
 * Ncurses user interface for Umix. Probably could use a total rewrite. But
 * hey, this works for me.
 *
 * Copyright (C) 2002 Sakari Lehtonen <sakari@ionstream.fi>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <ncurses.h>
#include "umix_common.h"
#include "ui.h"

#define UMIX_UI_NCURSES_VERSION "0.9"

/* the width of the channel labels */
static int labelwidth[MAX_MIXERS];	
/* store number of channels for every mixer globally for optimization */
static int numchan[MAX_MIXERS];
/* hack hackity, to be fixed soon */
static int ownupdate = 0;
/* the current mixer, stored here for optimization */
static int currmixer = 0;

/* position constants */
enum
{
	LABELMAX	= 10,		 /* max width of labels */
	BEGX		= 0,		 /* beginning x of the layout */
	BEGY		= 2,		 /* beginning y of the layout */
	VOLCHG		= (VOL_MAX / 50),/* amount of volume change */
	BALCHG		= 8,		 /* amount of balance change */
	NUMWIDTH	= 3,		 /* width of volume numbers */
};

/* position macros */

/* width of the bars */
#define BALWIDTH	((COLS/8) + 2)
#define VOLWIDTH	(COLS - BALWIDTH - NUMWIDTH - labelwidth[currmixer] - 8)

/* where the bars begin */
#define VOLBEGX		(BEGX+labelwidth[currmixer])
#define VOLBEGY		(BEGY+curr_chan)
#define BALBEGX		((COLS-3) - BALWIDTH)
#define BALBEGY		(BEGY+curr_chan)
#define STATBEGX	(VOLBEGX+NUMWIDTH)

/* the current channel marker positions */
#define VOLPOSX1	(VOLBEGX+NUMWIDTH+1)
#define VOLPOSX2	(VOLPOSX1+VOLWIDTH+1)
#define BALPOSX1	(BALBEGX-1)
#define BALPOSX2	(BALBEGX+BALWIDTH+1)

#ifndef CTRL
  #define CTRL(a)       ((a) & 0x1F) 
#endif

/* settings */
enum
{
	SET_LOAD,
	SET_SAVE,
	SET_CONFIG
};

/* which border */
enum
{
	BORD_UPPER,
	BORD_LOWER
};

/* mute action */
enum
{
	M_MUTE,
	M_EXCLUDE
};

/* the characters used in drawing */
enum
{
	CHAR_SLIDER	= '*',
	CHAR_DIV	= ':',
	CHAR_EMPTY	= '.',
	CHAR_MONO	= '|'
};

/* saved cols and lines */
static int saved_cols = 0;
static int saved_lines = 0;

/* global function prototypes */
int ui_ncurs_main(void);
void ui_ncurs_exit(void);
/*
plugin_info *ui_ncurs_get_info(void);
*/

/* internal ui_ncurses function prototypes */
static int main_loop(void);
static void initscreen(void);
static void cyclemixer(void);
static void mute(int action, int ch);
static void updatestat(void);
static void drawslider(void);
static void drawbalance(void);
static void drawstatus(void);
static void drawall(void);
static void drawborder(int where, const char *str);
static void drawpos(int menu, int status);
static void redrawscreen(void);
static void help(void);
static void about(void);
static void settings(int action);
static void input(char *buf, const char *prompt, int limit);
static void manualset(void);

ui_func ncurses_func =
{
	ui_ncurs_main,
	ui_ncurs_exit
		/*
	ui_ncurs_get_info
	*/
};

/*
plugin_info ncurses_info =
{
	"Ncurses UI",
	UMIX_UI_NCURSES_VERSION,
	UMIX_PLUGIN_UI_TEXT
};
*/

ui_func *get_ui_func(void)
{
	return &ncurses_func;
}

/*
plugin_info *get_plugin_info(void)
{
	return &ncurses_info;
}
*/

int ui_ncurs_main(void)
{
	int i, j, n;
	int numch;
	
	/* initialize curses */
	initscreen();

	/* calculate the max length of the labels */
	for(i=mixer_get_num_mix()-1; i>=0; i--)
	{
		mixer_set_curr(i);
		numch = mixer_get_num_chan();
		numchan[i] = numch;
		for(j=0; j<numch; j++)
		{
			chan_set_curr(j);
			n = strlen(chan_get_label());
			if( n > labelwidth[i] )
			{
				n = CLAMP(n, n, LABELMAX+1);
				labelwidth[i] = n;
			}
		}

	}
	mixer_set_curr(0);
	currmixer = 0;
	main_loop();
	
	return 0;
}

static int main_loop(void)
{
	int ch, curr_chan; 
	int menudir, curr_menu=0;
	int slidedir;
	float change; 
	int extradir;
	int quit=0;

	curr_chan = 0;
	chan_set_curr(curr_chan);
	while(1)
	{
		/* check if the window size has changed, this is called
		 * at least once in the beginning */
		if( (saved_cols != COLS) || (saved_lines != LINES) )
		{
			redrawscreen();
			drawpos(curr_menu, 1);
			saved_cols = COLS;
			saved_lines = LINES;
		}
		menudir = slidedir = extradir = 0;
		/* process input */
		if( (ch = getch()) != ERR )
		{
			switch(ch)
			{
			case 'j':
			case CTRL('n'):
			case KEY_DOWN:
				menudir = 1; break;
			case 'k': 
			case CTRL('p'):
			case KEY_UP:
				menudir = -1; break;
			case 'h':
			case CTRL('b'):
			case KEY_LEFT:
				slidedir = -1; break;
			case 'l':
			case CTRL('f'):
			case KEY_RIGHT:
				slidedir = 1; break;
			case ',':
			case '[':extradir = -(VOL_MAX/5); break;
			case '.':
			case ']':extradir = VOL_MAX/5; break;
			case ' ':chan_toggle_rec(); updatestat(); break;
			case CTRL('l'): redrawscreen(); break;
			case 'H':help(); break; 
			case 'A':about(); break;
			case 'L':settings(SET_LOAD); break;
			case 'S':settings(SET_SAVE); break;
			case 'e':manualset(); break;
			case 'f':settings(SET_CONFIG); break;
			case 'x':mute(M_EXCLUDE, curr_chan); 
				 break;
			case 'm':mute(M_MUTE, -1); break;
			case 'n':cyclemixer(); break;
			case 'q':quit = 1; break;
			default: break;
			}
			if( quit == 1 )
				break;
			
			/* yes hack */
			ownupdate = 1;
		}
		/* check if mixer has been updated */
		if( mixer_get_status() == 1 )
			/* if the current process didn't cause the update, 
			 * no need to draw again */
			if( ownupdate == 0 )
				drawall();
		ownupdate = 0;

		/* erase the old menu position */ 
		drawpos(curr_menu, 0);
		
		curr_chan += menudir;
		curr_chan = CLAMP(curr_chan, 0, numchan[currmixer]-1);
		/* this is just stupid .. TODO: implement scrolling of
		 * channels if too small to fit on screen */
		if( curr_chan > (LINES - 6) )
			curr_chan = LINES - 6;

		chan_set_curr(curr_chan);

		/* change the menu .. needs to be here for
		 * drawing the menupos */
		if( ch == '\t' )
			curr_menu = ~curr_menu;

		/* draw the menu position */
		drawpos(curr_menu, 1);
#ifdef UMIX_DEBUG
		mvprintw(20, 2, "loop: curr_chan=%2d curr_mixer=%2d", 
			curr_chan, currmixer);
#endif
		if( slidedir != 0 || extradir != 0 )
		{
			/* the volume menu */
			if( curr_menu == 0 )
			{
				change = (int)(slidedir * VOLCHG);
				change += (int) extradir;
				chan_inc_vol(change);
				drawslider();
			}	
			/* balance menu */
			else if( chan_is_stereo() )
			{
				if( extradir == 0 )
					change = 0.10 * slidedir;
				else
				{
					slidedir = (extradir > 1) ? 1:-1;
					change = 1.00 * slidedir;
				}
				chan_inc_bal(change);
				drawbalance();
			}
#ifdef UMIX_DEBUG
			mvprintw(23, 2, "loop: volume is %3d:%3d",
				chan_get_lr(CHAN_LEFT), 
				chan_get_lr(CHAN_RIGHT));
#endif
		}
		refresh();
		napms(20);
	}
	
	return 0;
}

/* update status for all channels */
static void updatestat(void)
{
	int i, orig_chan;

	orig_chan = chan_get_curr();
	for(i=0; i<numchan[currmixer]; i++)
	{
		chan_set_curr(i);
		drawstatus();
	}
	chan_set_curr(orig_chan);
}

/* cycles through available mixer devices */
static void cyclemixer(void)
{
	int old = currmixer;
	int num_mix = mixer_get_num_mix();

	if( currmixer+1 < num_mix )
		currmixer++;
	else if( currmixer+1 >= num_mix )
		currmixer = 0;

	if( currmixer != old )
	{
		mixer_set_curr(currmixer);
		redrawscreen();
	}
}

/* Mutes and demutes either all channels or one channel */
static void mute(int action, int ch)
{
	int i;
	int orig_chan;
	char *msg = NULL;

	orig_chan = chan_get_curr();
	/* all channels */
	if( ch == -1 )
	{
		for(i=0; i<numchan[currmixer]; i++)
		{
			chan_set_curr(i);
			/* if channel not already mute, mute */
			if( chan_is_mute() == 0 )
			{
				chan_set_mute(MUTE);
			}
			/* else restore old level */
			else
				chan_set_mute(UNMUTE);
			
			drawslider();
		}
		msg = _("Mute");
	}
	else
	{
		/* exclude channel */
		if( action == M_EXCLUDE )
		{
			/* mute */
			if( chan_is_mute() == 0 )
			{
				chan_set_mute(MUTE);
				msg = _("Exclude");
			}
			/* restore */
			else
			{
				chan_set_mute(UNMUTE);
				msg = _("Include");
			}
		}
		drawslider();
	}
	chan_set_curr(orig_chan);
	drawborder(BORD_LOWER, msg);
}

/* draws the menu position markers */
static void drawpos(int menu, int status)
{
	char start, end;
	int x1, x2;
	char buf[LABELMAX];
	int ch = chan_get_curr();

	if( status == 1 )
	{
		start = '>';
		end = '<';
	}
	else
		start = end = ' ';
	/* what menu are we in ? */
	if( menu == 0 )
	{
		x1 = VOLPOSX1;
		x2 = VOLPOSX2;		
	}
	else
	{
		x1 = BALPOSX1;
		x2 = BALPOSX2; 
	}
	
	attron(A_BOLD);
	mvaddch(BEGY+ch, x1, start);
	mvaddch(BEGY+ch, x2, end);
	attroff(A_BOLD);
	
	if( status == 1 )
		attron(A_BOLD);
	snprintf(buf, labelwidth[currmixer]+1, "%s", chan_get_label());
	mvaddstr(BEGY+ch, BEGX, buf);
	if( status == 1 ) 
		attroff(A_BOLD);
}

/* draws channel status information */
static void drawstatus(void)
{
	char rec;
	int ch = chan_get_curr();

	if( chan_is_record() )
	{
		if( chan_is_recsrc() )
			rec = 'R';
		else
			rec = 'P';
		
		attron(A_BOLD);
		mvprintw(BEGY+ch, STATBEGX, "%c", rec);
		attroff(A_BOLD);
	}
}

/* draws the volume level and slider */
static void drawslider(void)
{
	int i;
	int sliderwidth, max;
	int left, right;
	int y, x = VOLBEGX;

	y = BEGY + chan_get_curr();

	left = chan_get_lr(CHAN_LEFT);
	right = chan_get_lr(CHAN_RIGHT);
	max = MAX(left, right);
	sliderwidth = (max * VOLWIDTH) / VOL_MAX;
	mvprintw(y, x, "%3d", max);
	x += NUMWIDTH + 2;
	
	/* draw the characters indicating empty */
	for(i=0; i<VOLWIDTH; i++)
		mvaddch(y, x+i, CHAR_EMPTY);
	mvaddch(y, x, CHAR_DIV);
	mvaddch(y, x+(VOLWIDTH-1), CHAR_DIV);

	/* draw the slider */
	move(y, x);
	attron(A_BOLD);
	hline(CHAR_SLIDER, sliderwidth);
	attroff(A_BOLD);
}

/* draws the balance slider */
static void drawbalance(void)
{
	float bal;
	float pos;
	char mark;
	int ch;
	int y;
	int x = BALBEGX;

	ch = chan_get_curr();
	y = BEGY + ch;

	bal = chan_get_bal();
	
	if( bal == BAL_CENTER )
		pos = BALWIDTH/2;
	else
		pos = (BALWIDTH/2) + (bal * (BALWIDTH/2));

#ifdef UMIX_DEBUG
	mvprintw(26, 2, "drawbalance: pos=%2f %3f", 
		pos, bal);
#endif
	/* draw the base */
	move(y, x);
	hline(CHAR_EMPTY, BALWIDTH);
	/* draw the start, end and center markers */
	mvaddch(y, x, CHAR_DIV);
	mvaddch(y, x+BALWIDTH, CHAR_DIV);
	mvaddch(y, x+BALWIDTH/2, CHAR_DIV);
	
	if( chan_is_stereo() )
		mark = CHAR_SLIDER;
	else
		mark = CHAR_MONO;

	/* draw the marker */
	attron(A_BOLD);
	mvaddch(y, x+(int)pos, mark);
	attroff(A_BOLD);
}

/* draws the label, volume, balance slider and status for every channel */ 
static void drawall(void)
{
	int y = BEGY;
	int i;
	int origch;
	char buf[LABELMAX];

	origch = chan_get_curr();
	for(i=0; i<numchan[currmixer]; i++)
	{
		chan_set_curr(i);
		snprintf(buf, labelwidth[currmixer]+1, "%s", 
			chan_get_label());
		mvaddstr(y, BEGX, buf);
		drawslider();
		drawbalance();
		drawstatus();
		y++;
		/* TODO: fix when nothing better to do :) 
		 * kludge .. don't draw more volume bars than lines on 
		 * screen */
		if( y > (LINES - 4) )
			break;
	}
	chan_set_curr(origch);
}

/* draws the upper or lower border with optional string centered */
static void drawborder(int where, const char *str)
{
	int y, x;

	y = (where == BORD_UPPER ) ? 0 : LINES-2;

	attron(A_REVERSE);
	move(y, 0);
	hline(' ', COLS);
	if( str != NULL )
	{
		x = COLS/2 - strlen(str)/2;
		mvaddstr(y, x, str);
	}
	attroff(A_REVERSE);
}

/* refreshes the screen */
static void redrawscreen(void)
{
	char buf[LINEWIDTH]; 

	clear();
	snprintf(buf, sizeof(buf), "%s (%s::%s) %d/%d", get_version_str(), 
		mixer_get_name(), mixer_get_path(), mixer_get_curr()+1,
		mixer_get_num_mix());
	drawborder(BORD_UPPER, buf);
	snprintf(buf, sizeof(buf), _("Press H for help"));
	drawborder(BORD_LOWER, buf);
	drawall();
	refresh();
}

/* loads or saves settings depending on action */
static void settings(int action)
{
	int retval;
	char buf[BUFLEN];
	
	if( action == SET_LOAD )
	{
		retval = settings_load(globalopts.cfgpath, UMIX_SET_MIXER);
		if( retval == UMIX_ESETERR )
			snprintf(buf, sizeof(buf), "%s: %s",
				globalopts.cfgpath, strerror(errno));
		else if( retval == UMIX_ESETINVALID )
			snprintf(buf, sizeof(buf), UMIX_UI_EINVALID_STR);
		else if( retval == UMIX_ESETNOTFOUND )
			snprintf(buf, sizeof(buf), UMIX_UI_ENOTFOUND_STR);
		else if( retval == UMIX_ESETBINARY )
			snprintf(buf, sizeof(buf), UMIX_UI_ESETBINARY_STR);
		else
		{
			snprintf(buf, sizeof(buf), UMIX_UI_LOADOK_STRFMT,
				globalopts.cfgpath);
			drawall();
		}
	}
	else if( action == SET_SAVE )
	{
		if( (settings_save(globalopts.cfgpath) == UMIX_ESETERR ) )
			snprintf(buf, sizeof(buf), UMIX_UI_SAVEFAIL_STRFMT, 
				globalopts.cfgpath, strerror(errno));
		else
			snprintf(buf, sizeof(buf), UMIX_UI_SAVEOK_STRFMT, 
				globalopts.cfgpath);
	}
	else
	{
		/* falls to inputting the path for the config */

		int n;

		input(buf, _("Enter config path: "), 64);
		n = strlen(buf);
		if( n > 0 && n < COLS )
		{
			strncpy(globalopts.cfgpath, buf, 
				sizeof(globalopts.cfgpath));
			snprintf(buf, sizeof(buf), UMIX_UI_CFGSET_STRFMT,
				globalopts.cfgpath);
		}
		else
			snprintf(buf, sizeof(buf), UMIX_UI_CFGNOTSET_STRFMT,
				globalopts.cfgpath);

		drawborder(BORD_LOWER, buf);
		return;
	}
	/* draw the message */
	drawborder(BORD_LOWER, buf);
}

/* gets input from the keyboard to inbuf, with optional prompt */
static void input(char *inbuf, const char *prompt, int limit)
{
	if( limit <= 1 )
		return;
	nodelay(stdscr, FALSE);
	refresh();
	if( prompt != NULL )
		mvaddstr(LINES-1, 0, prompt); 	
	echo();
	getnstr(inbuf, limit-1);
	noecho();
	hline(' ', COLS);
	nodelay(stdscr, TRUE);
}

static void manualset()
{
	char inbuf[TINYSTRLEN/2];
	char buf[STRLEN];
	int retval = 0;
	int ch = chan_get_curr();
	char *str;

	snprintf(buf, sizeof(buf), _("Enter volume for channel %d: "), ch);
	input(inbuf, buf, sizeof(inbuf));
	if( inbuf[0] != '\0' )
	{
		retval = parsechanopt(inbuf);
		if( retval != 0 )
		{
			switch(retval)
			{
			case UMIX_EOPTLIMIT:
				str = UMIX_EOPTLIMIT_STR;
				break;
			case UMIX_EOPTINVALID:
				str = UMIX_EOPTINVALID_STR;
				break;
			case UMIX_EOPTNOTREC:
				str = UMIX_EOPTNOTREC_STR;
				break;
			default:
				str = inbuf;
				break;
			}
				
			snprintf(buf, sizeof(buf), UMIX_EOPTGENERAL_STRFMT, 
				chan_get_name(), str);
			drawborder(BORD_LOWER, buf);
		}
		else
		{
			drawborder(BORD_LOWER, NULL);
			drawslider();
			drawbalance();
			updatestat();
		}
	}
}

static void help(void)
{
	clear();
	drawborder(BORD_UPPER, _("Umix - Help"));
	drawborder(BORD_LOWER, _("Press any key to return"));
	mvaddstr(BEGY, BEGX,
		_(
		"j / ^n / arrow down\tMove down\n"
		"k / ^p / arrow up\tMove up\n"
		"h / ^b / arrow left\tDecrease volume / balance left\n"
		"l / ^f / arrow right\tIncrease volume / balance right\n"
		"] / ,\t\t\tIncrease volume by fifth of maximum amount\n"
		"[ / .\t\t\tDecrease volume by fifth of maximum amount\n"
		"Tab\t\t\tToggle between volume and balance menus\n"
		"Space\t\t\tToggle play / record source\n"
		"L\t\t\tLoad settings\n"
		"S\t\t\tSave settings\n"
		"f\t\t\tSpecify the config path for settings\n"
		"e\t\t\tEnter volume manually for current channel\n"
		"m\t\t\tMute all channels, undo if already mute\n"
		"n\t\t\tCycle through available mixers\n"
		"x\t\t\tSame as m but with only with the current channel\n"
		"H\t\t\tThis help\n"
		"A\t\t\tAbout Umix\n"
		"q\t\t\tExit the program\n"
		)
	);
	refresh();
	nodelay(stdscr, FALSE);
	getch();
	nodelay(stdscr, TRUE);
	redrawscreen();
}

static void about(void)
{
	char ver_buf[LINEWIDTH];

	snprintf(ver_buf, sizeof(ver_buf), 
		"%s - Ncurses UI v%s\n\n", get_version_str(),
		UMIX_UI_NCURSES_VERSION);

	clear();
	drawborder(BORD_UPPER, _("Umix - About"));
	drawborder(BORD_LOWER, _("Press any key to return"));
	mvaddstr(BEGY, BEGX, ver_buf);
	mvaddstr(BEGY+2, BEGX,
		_(
		"A program for adjusting soundcard volumes and other\n"
		"features in soundcard mixers.\n\n"
		"Copyright (C) 2002 Sakari Lehtonen <sakari@ionstream.fi>\n"
		)
	);
	refresh();
	nodelay(stdscr, FALSE);
	getch();
	nodelay(stdscr, TRUE);
	redrawscreen();
}

static void initscreen(void)
{
	initscr();
	keypad(stdscr, TRUE);
	nodelay(stdscr, TRUE);
	nonl();
	cbreak();
	noecho();
}

/* deinit ncurses */
void ui_ncurs_exit(void)
{
	clear();
	refresh();
	endwin();
#ifdef UMIX_DEBUG
	errormsg("ui_ncurs_exit: Ncurses UI deinitialized");
#endif 
}
