/*
 * umix_common.c: 
 * Common functions for Umix. All UIs have to be able to call
 * these, and errors are handled by the caller.
 * 
 * Copyright (C) 2002 Sakari Lehtonen <sakari@ionstream.fi>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "umix_common.h"

/* internal function prototypes */
static void parseglobal(char *buf);
static void parsemixer(char *buf);

/* internal variables */
static char *umix_version_str = NULL;

/* Parses one line of channel options.
 * 
 * Format is the following :
 *
 * vol|left:right[-|+]p|r
 * 
 * where 'name' is the name of the channel. 'vol' is either 
 * volume for both channels, or individual left and right values 
 * separated with a ':'. the last char sets the recording 
 * state of the channel to play 'p' or record source 'r'. 
 *
 * OR
 *
 * fromvol,tovol,noofsteps,delaytime p|r
 *
 * If you want to raise/decrease the volume with 'noofsteps' as the
 * number of steps from 'fromvol' to 'tovol', where 'delaytime' is the
 * delay between two steps.
 *
 * TODO: Okay, this starts to suck. Multiple sscanf etc, rewrite perhaps. 
 * */
int parsechanopt(const char *optarg)
{
	int change=0, slide=0;
	float left, right;
	volume vb;
	int recstate;
	size_t n;
	int retval = 0;

	/* skip whitespace */
	while(*optarg == ' ' || *optarg == '\t')
		optarg++;
	/* check for increment or decrement */
	n = strlen(optarg)-1;
	if( optarg[n] == '+' || optarg[n] == '-' )
		change = (optarg[n] == '+') ? 1 : -1;

#ifdef UMIX_DEBUG
	errormsg("parsechanopt: optarg=%s change=%d", optarg, change);
#endif
	/* if the first char is not a number, we might be setting only
	 * record sources */
	if( isdigit(optarg[0]) )
	{
		/* normal volume setting */
		if( strchr(optarg, ',') == NULL )
		{
			/* if no ':' found, use same value for both */
			if( strchr(optarg, ':') == NULL )
			{
				left = atof(optarg);
				right = left;
			}
			else
			{
				sscanf(optarg, "%f:%f", 
					&left, &right);
				if( chan_is_stereo() == 0 )
					right = left;
			}
		}
		/* increment the volume in steps */
		/* the format is this : fromvol,destvol,steps,delay */
		else
		{
			int i;
			float rinc, linc, ldir, rdir;
			float rdiff, ldiff;
			int steps = 20;
			long int delay = 20;
			float fleft, fright;
			volume fvb;

			slide = 1;
			if( strchr(optarg, ':') == NULL )
			{
				sscanf(optarg, "%f,%f,%d,%ld",
					&fleft, &left, &steps, &delay);
				fright = fleft;
				right = left;
			}
			else
			{
				sscanf(optarg, "%f:%f,%f:%f,%d,%ld",
					&fleft, &fright, 
					&left, &right,
					&steps, &delay);
			}
			/* we do not support steps going over the wanted
			 * left or right, because then increment would be
			 * zero */
			steps = CLAMP(steps, 0, VOL_MAX);
			ldiff = ABS(left - fleft);
			rdiff = ABS(right - fright);
			if( steps > ldiff && steps > rdiff )
				steps = MAX(ldiff, rdiff);

			/* check limits */
			if( (ldiff == 0 && rdiff == 0) || 
			    (ldiff > VOL_MAX && rdiff > VOL_MAX) ||steps == 0 )
				return UMIX_EOPTLIMIT;
			
			linc = left - fleft;
			if( linc < 0 )
			{
				ldir = -1;
				linc = ABS(linc);
			}
			else
				ldir = 1;
			linc /= steps;

			rinc = right - fright;
			if( rinc < 0 )
			{
				rdir = -1;
				rinc = ABS(rinc);
			}
			else
				rdir = 1;
			rinc /= steps;

			/* milliseconds, maximum is 5000 (5 seconds) */
			delay = CLAMP(delay, 0, 5000);
			delay *= 1000;
			/* correct direction */
			linc *= ldir;
			rinc *= rdir;
#ifdef UMIX_DEBUG
			errormsg("parsechanopt: ldiff=%3f rdiff=%3f "
				"steps=%d delay=%d", ldiff, rdiff,
				steps, delay);
#endif
			for(i=0; i<steps; i++)
			{
				lr_to_vb(fleft, fright, 
					&fvb.volume, &fvb.balance);
				chan_set_vb(fvb);
				fleft += linc;
				fright += rinc;
				usleep(delay);
			}
		}
	}
	else
	{
		/* check for garbage */
		recstate = tolower(optarg[0]);
		if( recstate != 'r' && recstate != 'p' )
			return UMIX_EOPTINVALID;
	}
	/* check for setting of record source or play */
	n = strcspn(optarg, "rRpP");
	recstate = tolower(optarg[n]);
	if( (recstate == 'r' || recstate == 'p') )
	{
		if( chan_is_record() )
		{
			if( recstate == 'r' )
				recstate = CHAN_RECSRC;
			else
				recstate = CHAN_PLAY;
			chan_set_rec(recstate);
		}
		else
			retval = UMIX_EOPTNOTREC;

		/* P or R is the only argument for this channel, return */
		if( isdigit(optarg[0]) == 0 )
			return retval;
	}
	/* check for overflowing */
	if( left > VOL_MAX || left < 0 || right > VOL_MAX || right < 0 )
		return UMIX_EOPTLIMIT;

	/* add or decrement from the old values ? if not sliding the vol */
	if( change != 0 && slide == 0 )
	{
		left = (change == 1 ) ? 
			chan_get_lr(CHAN_LEFT) + left : 
			chan_get_lr(CHAN_LEFT) - left;
		right = (change == 1 ) ? 
			chan_get_lr(CHAN_RIGHT) + right : 
			chan_get_lr(CHAN_RIGHT) - right;
	}
	
	lr_to_vb(left, right, &vb.volume, &vb.balance);
	chan_set_vb(vb);
#ifdef UMIX_DEBUG
	errormsg("parsechanopt: left=%d right=%d parsedleft=%f parsedright=%f",
		chan_get_lr(CHAN_LEFT), chan_get_lr(CHAN_RIGHT), 
		left, right);
#endif
	return retval;
}

/* the mixer that is used when parsing mixer settings */
static int set_mixer = -1;

/* load settings from file defined in path */
int settings_load(const char *path, int mode)
{
	FILE *fp;
	void (*parseline)(char *buf);
	char buf[LINEWIDTH];
	size_t n;
	int globalfound = 0, mixersfound = 0;

	/* lines are parsed depending on the mode */
	if( mode == UMIX_SET_GLOBAL )
		parseline = parseglobal;
	else
		parseline = parsemixer;

	if( path[0] == '-' && path[1] == '\0' )
		fp = stdin;
	else if( (fp = fopen(path, "r")) == NULL )
		return UMIX_ESETERR;

	while( fgets(buf, sizeof(buf), fp) != NULL )
	{
		/* try to detect binary files .. this _should_ detect 
		 * garbage/binary .. */
		if( isascii(buf[0]) == 0 )
			return UMIX_ESETBINARY;
		/* check for comments */
		n = strcspn(buf, "#");
		if( n < 2 )
			continue;

		/* remove newline */
		n = strlen(buf) - 1;
		if( buf[n] == '\n' || buf[n] == '\r' )
			buf[n] = '\0';

		/* check for setting sections */
		if( strncmp(buf, "[global]", 8) == 0 )
		{
			globalfound = 1;
			if( (mode == UMIX_SET_MIXER) && (mixersfound == 1) )
				break;
			continue;
		}
		else if( strncmp(buf, "[mixers]", 8) == 0 )
		{
			mixersfound = 1;
			if( (mode == UMIX_SET_GLOBAL) && (globalfound == 1) )
				break;
			continue;
		}
		/* parse lines if the mode is right and beginning found */
		if( ((mode == UMIX_SET_GLOBAL) && globalfound == 1) ||
		    ((mode == UMIX_SET_MIXER) && mixersfound == 1 ))
			parseline(buf);
	}

	/* no global or mixers section found, assume invalid config file */
	if( globalfound == 0 && mixersfound == 0 )
		return UMIX_ESETINVALID;

	/* mixer settings for specified mixer not found */
	if( (mode == UMIX_SET_MIXER) && (set_mixer == -1) )
		return UMIX_ESETNOTFOUND;

	if( fp == stdin )
	 	return 0;
	else fclose(fp);

	return 0;
}

/* parses one line in the [global] section of config file */
static void parseglobal(char *buf)
{
	size_t n;

	/* hmm yes there might be a better way to do this .. but this is
	 * sufficient and simple enough .. not anymore :P */
	n = strcspn(buf, "=");
	buf[n++] = '\0';
	if( strcmp(buf, "device") == 0 )
		strncpy(globalopts.devpath, buf + n,
			sizeof(globalopts.devpath));
	else if( strcmp(buf, "config") == 0 )
		strncpy(globalopts.cfgpath, buf + n,
			sizeof(globalopts.cfgpath));
	else if( strcmp(buf, "ui") == 0 )
		strncpy(globalopts.ui, buf + n,
			sizeof(globalopts.ui));
	else if( strcmp(buf, "driver") == 0 )
		strncpy(globalopts.driver, buf + n,
			sizeof(globalopts.driver));
	else if( strcmp(buf, "plugindir") == 0 )
		strncpy(globalopts.plugindir, buf + n,
			sizeof(globalopts.plugindir));
}

/* parses one line in the [mixers] section of config file */
static void parsemixer(char *buf)
{
	size_t n;
	int i;

	/* these aren't really required in the config file .. they are
	 * just helping readability */
	if( *buf == '{' || *buf == '}' )
		return;
	/* jump over leading whitespace */
	while( *buf == ' ' || *buf == '\t' )
		buf++;

	/* find the mixer  */
	if( strncmp(buf, "mixer=", 6) == 0 )
	{
		buf += 6;
		for(i=mixer_get_num_mix()-1; i>=0; i--)
		{
			/* compare device paths to find out the right 
			 * mixer */
			mixer_set_curr(i);
			if( strcmp(buf, mixer_get_path()) == 0 )
			{
				/* found */
				set_mixer = i;
				break;
			}
		}
	}
	/* mixer already found */
	else if( set_mixer != -1 )
	{
		int chnum;

		mixer_set_curr(set_mixer);
		n = strcspn(buf, " ");
		buf[n] = '\0';
		/* parse the channel option */
		if( (chnum = mixer_opt_to_chan_num(buf)) != -1 )
		{
			chan_set_curr(chnum);
			parsechanopt(buf+n+1);
		}
	}
}

/* save mixer settings to a file defined in path. */
int settings_save(const char *path)
{
	FILE *fp;
	int i, j, n;
	char record;

	if( path[0] == '-' && path[1] == '\0' )
		fp = stdout;
	else if( (fp = fopen(path, "w+")) == NULL )
		return UMIX_ESETERR;

	fprintf(fp, "# Umix config file\n"
		    "[global]\n"
		    "device=%s\n"
		    "config=%s\n"
		    "plugindir=%s\n"
		    "ui=%s\n"
		    "driver=%s\n\n",
		    globalopts.devpath, globalopts.cfgpath,
		    globalopts.plugindir, globalopts.ui, globalopts.driver);
	fprintf(fp, "[mixers]\n");
	
	/* hmhm. we do this so we dont have to call mixer_get_num_mix()
	 * over every iteration. */
	for(i=mixer_get_num_mix()-1; i>=0; i--)
	{
		mixer_set_curr(i);
		/* print the name of the mixer and mixer driver name
		 * just to add readability */
		fprintf(fp, "# %s, %s\n", mixer_get_name(), mixer_get_driver());
		fprintf(fp, "mixer=%s\n{\n", mixer_get_path());

		/* we must do this to get the channels in right order */
		n = mixer_get_num_chan();
		for(j=0; j<n; j++)
		{
			chan_set_curr(j);
			if( chan_is_recsrc() )
				record = 'R';
			else if( chan_is_record() )
				record = 'P';
			else
				record = ' ';
			fprintf(fp, "\t%s %d:%d %c\n", 
				chan_get_name(), chan_get_lr(CHAN_LEFT),
				chan_get_lr(CHAN_RIGHT), record);
		}
		fprintf(fp, "}\n");
	}
	
	/* we dont want to close stdout .. */
	if( fp == stdout )
		return 0;
	else fclose(fp);

	return 0;
}

void init_version(void)
{
	char buf[TINYSTRLEN];

	snprintf(buf, sizeof(buf), "%s v%s", NAME_STR, VERSION);
	if( umix_version_str == NULL )
		umix_version_str = strdup(buf);
}

char *get_version_str(void)
{
	return umix_version_str;
}

/* our very simple growing array code */
/* add nameval to growing array tab */
int nv_add(nvtab *tab, nameval newname)
{
	nameval *nvp;

	/* first time called on tab */
	if( tab->nv == NULL )
	{	
		tab->nv = malloc(NVINIT * sizeof(nameval));
		if( tab->nv == NULL )
			return -1;
		tab->max = NVINIT;
		tab->nval = 0;
	}
	else if( tab->nval >= tab->max )
	{
		/* reallocate nv by multiplier of NVGROW */
		nvp = erealloc(tab->nv, (NVGROW*tab->max) * sizeof(nameval));
		if( nvp == NULL )
			return -1;
		tab->max *= NVGROW;
		tab->nv = nvp;
	}
	tab->nv[tab->nval] = newname;
	
	return tab->nval++;
}

/* search and return nameval by name from tab */
void *nv_lookupbyname(nvtab *tab, const char *name)
{
	int i;

	for(i=0; i<tab->nval; i++)
	{
		if( strcmp(tab->nv[i].name, name) == 0 )
			return tab->nv[i].val;
	}

	return NULL;
}
